import time
import os.path as path


modules = {}
names = ('misaka', 'markdown', 'markdown2', 'cMarkdown', 'discount')
for name in names:
    try:
        modules[name] = __import__(name)
    except ImportError:
        pass


rndr = modules['misaka'].HtmlRenderer()
m = modules['misaka'].Markdown(rndr)


class Benchmark(object):
    def __init__(self, name):
        self._name = name

    def __call__(self, func):
        def wrapper(*args, **kwargs):
            start = time.clock()
            func(*args, **kwargs)
            end = time.clock()
            return end - start
        wrapper.__name__ = func.__name__
        return wrapper


@Benchmark('Misaka')
def benchmark_misaka(text):
    modules['misaka'].html(text)


@Benchmark('Misaka (classes)')
def benchmark_misaka_classes(text):
    m.render(text)


@Benchmark('markdown2')
def benchmark_markdown2(text):
    modules['markdown2'].markdown(text)


@Benchmark('Markdown')
def benchmark_markdown(text):
    modules['markdown'].markdown(text)


@Benchmark('cMarkdown')
def benchmark_cMarkdown(text):
    modules['cMarkdown'].markdown(text)


@Benchmark('discount')
def benchmark_discount(text):
    modules['discount'].Markdown(text).get_html_content()


if __name__ == '__main__':
    with open(path.join(path.dirname(__file__), 'markdown-syntax.md'), 'r') as fd:
        text = fd.read()

    loops = 10000
    totals = []
    methods = [
        ('Misaka', benchmark_misaka),
        ('Misaka (classes)', benchmark_misaka_classes),
        ('Markdown', benchmark_markdown),
        ('Markdown2', benchmark_markdown2),
        ('cMarkdown', benchmark_cMarkdown),
        ('Discount', benchmark_discount)
    ]

    print('Parsing the Markdown Syntax document %d times...' % loops)

    for i, method in enumerate(methods):
        name = method[1].__name__.split('_', 2)[1]
        if name not in modules:
            print('%s is not available' % method[0])
            continue
        total = 0
        for nth in range(0, loops):
            total += method[1](text)
        print('%s: %gs' % (method[0], total))



# -*- coding: utf-8 -*-

# Most unit tests are based on the tests from Redcarpet.

import re
import codecs
from os import path
from glob import glob
from subprocess import Popen, PIPE, STDOUT

import misaka

from misaka import Markdown, BaseRenderer, HtmlRenderer, SmartyPants, \
    EXT_NO_INTRA_EMPHASIS, EXT_TABLES, EXT_FENCED_CODE, EXT_AUTOLINK, \
    EXT_STRIKETHROUGH, EXT_LAX_HTML_BLOCKS, EXT_SPACE_HEADERS, \
    EXT_SUPERSCRIPT, \
    HTML_SKIP_HTML, HTML_SKIP_STYLE, HTML_SKIP_IMAGES, HTML_SKIP_LINKS, \
    HTML_EXPAND_TABS, HTML_SAFELINK, HTML_TOC, HTML_HARD_WRAP, \
    HTML_USE_XHTML, HTML_ESCAPE, \
    HTML_SMARTYPANTS

from minitest import TestCase, ok, runner


def clean_html(dirty_html):
    input_html = dirty_html.encode('utf-8')
    p = Popen(['tidy', '--show-body-only', '1', '--quiet', '1', '--show-warnings', '0', '-utf8'],
        stdout=PIPE, stdin=PIPE, stderr=STDOUT)
    stdout, stderr = p.communicate(input=input_html)

    return stdout.decode('utf-8')


class SmartyPantsTest(TestCase):
    name = 'SmartyPants'

    def setup(self):
        self.r = lambda html: misaka.html(html, render_flags=HTML_SMARTYPANTS)

    def test_single_quotes_re(self):
        html = self.r('<p>They\'re not for sale.</p>\n')
        ok(html).diff('<p>They&rsquo;re not for sale.</p>\n')

    def test_single_quotes_ll(self):
        html = self.r('<p>Well that\'ll be the day</p>\n')
        ok(html).diff('<p>Well that&rsquo;ll be the day</p>\n')

    def test_double_quotes_to_curly_quotes(self):
        html = self.r('<p>"Quoted text"</p>\n')
        ok(html).diff('<p>&ldquo;Quoted text&rdquo;</p>\n')

    def test_single_quotes_ve(self):
        html = self.r('<p>I\'ve been meaning to tell you ..</p>\n')
        ok(html).diff('<p>I&rsquo;ve been meaning to tell you ..</p>\n')

    def test_single_quotes_m(self):
        html = self.r('<p>I\'m not kidding</p>\n')
        ok(html).diff('<p>I&rsquo;m not kidding</p>\n')

    def test_single_quotes_d(self):
        html = self.r('<p>what\'d you say?</p>\n')
        ok(html).diff('<p>what&rsquo;d you say?</p>\n')


class HtmlRenderTest(TestCase):
    name = 'Html Renderer'

    def setup(self):
        pants = SmartyPants()

        self.r = {
            HTML_SKIP_HTML: HtmlRenderer(HTML_SKIP_HTML),
            HTML_SKIP_IMAGES: HtmlRenderer(HTML_SKIP_IMAGES),
            HTML_SKIP_LINKS: HtmlRenderer(HTML_SKIP_LINKS),
            HTML_SAFELINK: HtmlRenderer(HTML_SAFELINK),
            HTML_ESCAPE: HtmlRenderer(HTML_ESCAPE),
            HTML_HARD_WRAP: HtmlRenderer(HTML_HARD_WRAP)
        }

    def render_with(self, flag, text):
        return Markdown(self.r[flag]).render(text)

    # Hint: overrides HTML_SKIP_HTML, HTML_SKIP_IMAGES and HTML_SKIP_LINKS
    def test_escape_html(self):
        source = '''
Through <em>NO</em> <script>DOUBLE NO</script>

<script>BAD</script>

<img src="/favicon.ico" />
'''

        expected = clean_html('''
<p>Through &lt;em&gt;NO&lt;/em&gt; &lt;script&gt;DOUBLE NO&lt;/script&gt;</p>

<p>&lt;script&gt;BAD&lt;/script&gt;</p>

<p>&lt;img src=&quot;/favicon.ico&quot; /&gt;</p>
''')

        markdown = clean_html(self.render_with(HTML_ESCAPE, source))
        ok(markdown).diff(expected)

    def test_skip_html(self):
        markdown = self.render_with(HTML_SKIP_HTML, 'Through <em>NO</em> <script>DOUBLE NO</script>')
        ok(markdown).diff('<p>Through NO DOUBLE NO</p>\n')

    def test_skip_html_two_space_hard_break(self):
        markdown = self.render_with(HTML_SKIP_HTML, 'Lorem,  \nipsum\n')
        ok(markdown).diff('<p>Lorem,<br>\nipsum</p>\n')

    def test_skip_image(self):
        markdown = self.render_with(HTML_SKIP_IMAGES, '![dust mite](http://dust.mite/image.png) <img src="image.png" />')
        ok(markdown).not_contains('<img')

    def test_skip_links(self):
        markdown = self.render_with(HTML_SKIP_LINKS, '[This link](http://example.net/) <a href="links.html">links</a>')
        ok(markdown).not_contains('<a ')

    def test_safelink(self):
        markdown = self.render_with(HTML_SAFELINK, '[IRC](irc://chat.freenode.org/#freenode)')
        ok(markdown).diff('<p>[IRC](irc://chat.freenode.org/#freenode)</p>\n')

    def test_hard_wrap(self):
        markdown = self.render_with(HTML_HARD_WRAP, '''
Hello world,
this is just a simple test

With hard wraps
and other *things*.''')
        ok(markdown).contains('<br>')


class MarkdownParserTest(TestCase):
    name = 'Markdown Parser'

    def setup(self):
        self.r = Markdown(HtmlRenderer()).render

    def render_with(self, text, flags=0, extensions=0):
        return Markdown(HtmlRenderer(), extensions).render(text)

    def test_one_liner_to_html(self):
        markdown = self.r('Hello World.')
        ok(markdown).diff('<p>Hello World.</p>\n')

    def test_inline_markdown_to_html(self):
        markdown = self.r('_Hello World_!')
        ok(markdown).diff('<p><em>Hello World</em>!</p>\n')

    def test_inline_markdown_start_end(self):
        markdown = self.render_with('_start _ foo_bar bar_baz _ end_ *italic* **bold** <a>_blah_</a>',
            extensions=EXT_NO_INTRA_EMPHASIS)
        ok(markdown).diff('<p><em>start _ foo_bar bar_baz _ end</em> <em>italic</em> <strong>bold</strong> <a><em>blah</em></a></p>\n')

        markdown = self.r('Run \'rake radiant:extensions:rbac_base:migrate\'')
        ok(markdown).diff('<p>Run &#39;rake radiant:extensions:rbac_base:migrate&#39;</p>\n')

    def test_urls_not_doubly_escaped(self):
        markdown = self.r('[Page 2](/search?query=Markdown+Test&page=2)')
        ok(markdown).diff('<p><a href="/search?query=Markdown+Test&amp;page=2">Page 2</a></p>\n')

    def test_inline_html(self):
        markdown = self.r('before\n\n<div>\n  foo\n</div>\n\nafter')
        ok(markdown).diff('<p>before</p>\n\n<div>\n  foo\n</div>\n\n<p>after</p>\n')

    def test_html_block_end_tag_on_same_line(self):
        markdown = self.r('Para 1\n\n<div><pre>HTML block\n</pre></div>\n\nPara 2 [Link](#anchor)')
        ok(markdown).diff('<p>Para 1</p>\n\n<div><pre>HTML block\n</pre></div>\n\n<p>Para 2 <a href=\"#anchor\">Link</a></p>\n')

    # This isn't in the spec but is Markdown.pl behavior.
    def test_block_quotes_preceded_by_spaces(self):
        markdown = self.r(
            'A wise man once said:\n\n' \
            ' > Isn\'t it wonderful just to be alive.\n')
        ok(markdown).diff(
            '<p>A wise man once said:</p>\n\n' \
            '<blockquote>\n<p>Isn&#39;t it wonderful just to be alive.</p>\n</blockquote>\n')

    def test_html_block_not_wrapped_in_p(self):
        markdown = self.render_with(
            'Things to watch out for\n' \
            '<ul>\n<li>Blah</li>\n</ul>\n',
            extensions=EXT_LAX_HTML_BLOCKS)
        ok(markdown).diff(
            '<p>Things to watch out for</p>\n\n' \
            '<ul>\n<li>Blah</li>\n</ul>\n')

    # http://github.com/rtomayko/rdiscount/issues/#issue/13
    def test_headings_with_trailing_space(self):
        markdown = self.render_with(
            'The Ant-Sugar Tales \n' \
            '=================== \n\n' \
            'By Candice Yellowflower   \n',
            extensions=EXT_LAX_HTML_BLOCKS)
        ok(markdown).diff('<h1>The Ant-Sugar Tales </h1>\n\n<p>By Candice Yellowflower   </p>\n')

    def test_intra_emphasis(self):
        markdown = self.r('foo_bar_baz')
        ok(markdown).diff('<p>foo<em>bar</em>baz</p>\n')

        markdown = self.render_with('foo_bar_baz', extensions=EXT_NO_INTRA_EMPHASIS)
        ok(markdown).diff('<p>foo_bar_baz</p>\n')

    def test_autolink(self):
        markdown = self.render_with('http://axr.vg/', extensions=EXT_AUTOLINK)
        ok(markdown).diff('<p><a href=\"http://axr.vg/\">http://axr.vg/</a></p>\n')

    def test_tags_with_dashes_and_underscored(self):
        markdown = self.r('foo <asdf-qwerty>bar</asdf-qwerty> and <a_b>baz</a_b>')
        ok(markdown).diff('<p>foo <asdf-qwerty>bar</asdf-qwerty> and <a_b>baz</a_b></p>\n')

    def test_no_link_in_code_blocks(self):
        markdown = self.r('    This is a code block\n    This is a link [[1]] inside\n')
        ok(markdown).diff('<pre><code>This is a code block\nThis is a link [[1]] inside\n</code></pre>\n')

    def test_whitespace_after_urls(self):
        markdown = self.render_with('Japan: http://www.abc.net.au/news/events/japan-quake-2011/beforeafter.htm (yes, japan)',
            extensions=EXT_AUTOLINK)
        ok(markdown).diff('<p>Japan: <a href="http://www.abc.net.au/news/events/japan-quake-2011/beforeafter.htm">http://www.abc.net.au/news/events/japan-quake-2011/beforeafter.htm</a> (yes, japan)</p>\n')

    def test_infinite_loop_in_header(self):
        markdown = self.render_with(
            '######\n' \
            '#Body#\n' \
            '######\n',
            extensions=EXT_LAX_HTML_BLOCKS)
        ok(markdown).diff('<h1>Body</h1>\n')

    def test_tables(self):
        text = ' aaa | bbbb\n' \
            '-----|------\n' \
            'hello|sailor\n'

        ok(self.render_with(text)).not_contains('<table')
        ok(self.render_with(text, extensions=EXT_TABLES)).contains('<table')

    def test_strikethrough(self):
        text = 'this is ~some~ striked ~~text~~'

        ok(self.render_with(text)).not_contains('<del')
        ok(self.render_with(text, extensions=EXT_STRIKETHROUGH)).contains('<del')

    def test_fenced_code_blocks(self):
        text = '''
This is a simple test

~~~~~
This is some awesome code
    with tabs and shit
~~~
'''

        ok(self.render_with(text)).not_contains('<code')
        ok(self.render_with(text, extensions=EXT_FENCED_CODE)).contains('<code')

    def test_linkable_headers(self):
        markdown = self.r('### Hello [GitHub](http://github.com)')
        ok(markdown).diff('<h3>Hello <a href=\"http://github.com\">GitHub</a></h3>\n')

    def test_autolinking_with_ent_chars(self):
        markdown = self.render_with('This a stupid link: https://github.com/rtomayko/tilt/issues?milestone=1&state=open',
            extensions=EXT_AUTOLINK)
        ok(markdown).diff('<p>This a stupid link: <a href=\"https://github.com/rtomayko/tilt/issues?milestone=1&amp;state=open\">https://github.com/rtomayko/tilt/issues?milestone=1&amp;state=open</a></p>\n')

    def test_spaced_headers(self):
        text = '#123 a header yes\n'
        ok(self.render_with(text, extensions=EXT_SPACE_HEADERS)).not_contains('<h1>')


class MarkdownConformanceTest_10(TestCase):
    name = 'Markdown Conformance 1.0'
    suite = 'MarkdownTest_1.0'

    def setup(self):
        self.r = Markdown(HtmlRenderer()).render

        tests_dir = path.dirname(__file__)
        for text_path in glob(path.join(tests_dir, self.suite, '*.text')):
            html_path = '%s.html' % path.splitext(text_path)[0]
            self._create_test(text_path, html_path)

    def _create_test(self, text_path, html_path):
        def test():
            with codecs.open(text_path, 'r', encoding='utf-8') as fd:
                text = fd.read()
            with codecs.open(html_path, 'r', encoding='utf-8') as fd:
                expected_html = fd.read()

            actual_html = self.r(text)
            expected_result = clean_html(expected_html)
            actual_result = clean_html(actual_html)

            ok(actual_result).diff(expected_result)

        test.__name__ = self._test_name(text_path)
        self.add_test(test)

    def _test_name(self, text_path):
        name = path.splitext(path.basename(text_path))[0]
        name = name.replace(' - ', '_')
        name = name.replace(' ', '_')
        name = re.sub('[(),]', '', name)
        return 'test_%s' % name.lower()


class MarkdownConformanceTest_103(MarkdownConformanceTest_10):
    name = 'Markdown Conformance 1.0.3'
    suite = 'MarkdownTest_1.0.3'


class UnicodeTest(TestCase):
    name = 'Unicode'

    def setup(self):
        self.r = Markdown(HtmlRenderer()).render

    def test_unicode(self):
        tests_dir = path.dirname(__file__)

        with codecs.open(path.join(tests_dir, 'unicode.txt'), 'r', encoding='utf-8') as fd:
            text = fd.read()
        with codecs.open(path.join(tests_dir, 'unicode.html'), 'r', encoding='utf-8') as fd:
            html = fd.read()

        markdown = self.r(text)
        ok(markdown).diff(html)


if __name__ == '__main__':
    runner([
        SmartyPantsTest,
        HtmlRenderTest,
        MarkdownParserTest,
        MarkdownConformanceTest_10,
        MarkdownConformanceTest_103,
        UnicodeTest
    ])



# minitest is inspired by Oktest, http://www.kuwata-lab.com/oktest/

import re
import sys
from difflib import unified_diff
from collections import namedtuple


_RE_TEST_NAME = re.compile(r'test_[a-zA-Z0-9_]+')
Result = namedtuple('Result', ('func', 'name', 'error'))


def msg(message):
    return '{0}\n{1}\n{0}'.format('-' * 80, message)


class ExtendedDict(dict):
    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            return None

    def __setattr__(self, name, value):
        self[name] = value


class AssertionObject(object):
    def __init__(self, target):
        self._target = target

    def __lt__(self, other):
        if not self._target < other:
            raise AssertionError(msg('%r < %r' % (self._target, other)))

    def __le__(self, other):
        if not self._target <= other:
            raise AssertionError(msg('%r <= %r' % (self._target, other)))


    def __eq__(self, other):
        if not self._target == other:
            raise AssertionError(msg('%r == %r' % (self._target, other)))

    def __ne__(self, other):
        if not self._target != other:
            raise AssertionError(msg('%r != % ' % (self._target, other)))

    def __gt__(self, other):
        if not self._target > other:
            raise AssertionError(msg('%r > %r' % (self._target, other)))

    def __ge__(self, other):
        if not self._target >= other:
            raise AssertionError(msg('%r >= %r' % (self._target, other)))

    def diff(self, other):
        if self._target != other:
            difference = unified_diff(
                other.splitlines(True),
                self._target.splitlines(True))
            raise AssertionError(msg(''.join(difference)))

    def contains(self, other):
        if other not in self._target:
            raise AssertionError(msg('%r in %r' % (other, self._target)))

    def not_contains(self, other):
        if other in self._target:
            raise AssertionError(msg('%r not in %r' % (other, self._target)))


def ok(target):
    return AssertionObject(target)


class TestCase(object):
    def __init__(self, config):
        self.config = config
        self._tests = []
        for t in dir(self):
            if _RE_TEST_NAME.match(t):
                self.add_test(getattr(self, t))

    def add_test(self, func):
        def catch_exception():
            try:
                func()
                error = None
            except AssertionError as e:
                error = str(e)
            return Result(func.__name__, func.__doc__, error)
        self._tests.append(catch_exception)
        return catch_exception

    def setup(self):
        pass

    def teardown(self):
        pass

    def run(self):
        self.setup()
        for test in self._tests:
            yield test()
        self.teardown()


def runner(testcases, config={}):
    passed = failed = 0
    config = ExtendedDict(config)

    for testcase in testcases:
        tests = testcase(config)

        if hasattr(tests, 'name'):
            print('\n>> %s' % tests.name)

        for result in tests.run():
            name = result.name or result.func
            if result.error is not None:
                failed += 1
                print('%s ... FAILED\n\n%s\n' % (name, result.error))
            else:
                passed += 1
                print('%s ... PASSED' % name)

    print('\n\n%s passed; %s failed.' % (passed, failed))



# -*- coding: utf-8 -*-

from misaka import Markdown, BaseRenderer, HtmlRenderer, \
    SmartyPants, \
    EXT_FENCED_CODE, EXT_TABLES, EXT_AUTOLINK, EXT_STRIKETHROUGH, \
    EXT_SUPERSCRIPT, HTML_USE_XHTML, \
    TABLE_ALIGN_L, TABLE_ALIGN_R, TABLE_ALIGN_C, \
    TABLE_ALIGNMASK, TABLE_HEADER


class BleepRenderer(HtmlRenderer, SmartyPants):

    def block_code(self, text, lang):
        if lang:
            lang = ' class="%s"' % lang
        else:
            lang  = ''
        return '\n<pre%s><code>%s</code></pre>\n' % (lang, text)

    def block_quote(self, text):
        return '\n<blockquote>%s</blockquote>\n' % text

    def block_html(self, text):
        return '\n%s' % text

    def header(self, text, level):
        return '\n<h%d>%s</h%d>\n' % (level, text, level)

    def hrule(self):
        if self.flags & HTML_USE_XHTML:
            return '\n<hr/>\n'
        else:
            return '\n<hr>\n'

    def list(self, text, is_ordered):
        if is_ordered:
            return '\n<ol>%s</ol>\n' % text
        else:
            return '\n<ul>%s</ul>\n' % text

    def list_item(self, text, is_ordered):
        return '<li>%s</li>\n' % text

    def paragraph(self, text):
        # No hard wrapping yet. Maybe with:
        # http://docs.python.org/library/textwrap.html
        return '\n<p>%s</p>\n' % text

    def table(self, header, body):
        return '\n<table><thead>\n%s</thead><tbody>\n%s</tbody></table>\n' % \
            (header, body)

    def table_row(self, text):
        return '<tr>\n%s</tr>\n' % text

    def table_cell(self, text, flags):
        flags = flags & TABLE_ALIGNMASK
        if flags == TABLE_ALIGN_C:
            align = 'align="center"'
        elif flags == TABLE_ALIGN_L:
            align = 'align="left"'
        elif flags == TABLE_ALIGN_R:
            align = 'align="right"'
        else:
            align = ''

        if flags & TABLE_HEADER:
            return '<th%s>%s</th>\n' % (align, text)
        else:
            return '<td%s>%s</td>\n' % (align, text)

    def autolink(self, link, is_email):
        if is_email:
            return '<a href="mailto:%(link)s">%(link)s</a>' % {'link': link}
        else:
            return '<a href="%(link)s">%(link)s</a>' % {'link': link}

    def preprocess(self, text):
        return text.replace(' ', '_')


md = Markdown(BleepRenderer(),
    EXT_FENCED_CODE | EXT_TABLES | EXT_AUTOLINK |
    EXT_STRIKETHROUGH | EXT_SUPERSCRIPT)


print(md.render('''
Unordered

- One
- Two
- Three

And now ordered:

1. Three
2. Two
3. One

An email: example@example.com

And an URL: http://example.com
'''))



cimport sundown
cimport wrapper

from libc.stdint cimport uint8_t


__version__ = '1.0.2'

# Markdown extensions
EXT_NO_INTRA_EMPHASIS = (1 << 0)
EXT_TABLES = (1 << 1)
EXT_FENCED_CODE = (1 << 2)
EXT_AUTOLINK = (1 << 3)
EXT_STRIKETHROUGH = (1 << 4)
EXT_LAX_HTML_BLOCKS = (1 << 5)
EXT_SPACE_HEADERS = (1 << 6)
EXT_SUPERSCRIPT = (1 << 7)

# HTML Render flags
HTML_SKIP_HTML = (1 << 0)
HTML_SKIP_STYLE = (1 << 1)
HTML_SKIP_IMAGES = (1 << 2)
HTML_SKIP_LINKS = (1 << 3)
HTML_EXPAND_TABS = (1 << 4)
HTML_SAFELINK = (1 << 5)
HTML_TOC = (1 << 6)
HTML_HARD_WRAP = (1 << 7)
HTML_USE_XHTML = (1 << 8)
HTML_ESCAPE = (1 << 9)

# Extra HTML render flags - these are not from Sundown
HTML_SMARTYPANTS = (1 << 10)  # An extra flag to enable Smartypants
HTML_TOC_TREE = (1 << 11)  # Only render a table of contents tree

# Other flags
TABLE_ALIGN_L = 1 # MKD_TABLE_ALIGN_L
TABLE_ALIGN_R = 2 # MKD_TABLE_ALIGN_R
TABLE_ALIGN_C = 3 # MKD_TABLE_ALIGN_CENTER
TABLE_ALIGNMASK = 3 # MKD_TABLE_ALIGNMASK
TABLE_HEADER = 4 # MKD_TABLE_HEADER


def html(object text, unsigned int extensions=0, unsigned int render_flags=0):
    """Convert markdown text to (X)HTML.

    Returns a unicode string.

    :param text: A byte or unicode string.
    :param extensions: Enable additional Markdown extensions with the ``EXT_*`` constants.
    :param render_flags: Adjust HTML rendering behaviour with the ``HTML_*`` constants.
    """

    if render_flags & HTML_TOC_TREE:
        renderer = HtmlTocRenderer(render_flags)
    else:
        renderer = HtmlRenderer(render_flags)

    markdown = Markdown(renderer, extensions)
    result = markdown.render(text)

    if render_flags & HTML_SMARTYPANTS:
        result = SmartyPants().postprocess(result)

    return result


cdef class SmartyPants:
    """Smartypants is a post-processor for (X)HTML renderers and can be used
    standalone or as a mixin. It adds a methode named ``postprocess`` to the
    renderer.

    ================================== ========
    Source                             Result
    ================================== ========
    `'s` (s, t, m, d, re, ll, ve) [1]_ &rsquo;s
    `--`                               &mdash;
    `-`                                &ndash;
    `...`                              &hellip;
    `. . .`                            &hellip;
    `(c)`                              &copy;
    `(r)`                              &reg;
    `(tm)`                             &trade;
    `3/4`                              &frac34;
    `1/2`                              &frac12;
    `1/4`                              &frac14;
    ================================== ========

    .. [1] A ``'`` followed by a ``s``, ``t``, ``m``, ``d``, ``re``, ``ll`` or
           ``ve`` will be turned into ``&rsquo;s``, ``&rsquo;t``, and so on.
    """
    def postprocess(self, object text):
        """Process the input text.

        Returns a unicode string.

        :param text: A byte or unicode string.
        """
        # Convert string
        cdef bytes py_string
        if hasattr(text, 'encode'):
            py_string = text.encode('UTF-8', 'strict')
        else:
            py_string = text
        cdef char *c_string = py_string

        cdef sundown.buf *ob = sundown.bufnew(128)
        sundown.sdhtml_smartypants(ob,
            <uint8_t *> c_string, len(c_string))

        try:
            return (<char *> ob.data)[:ob.size].decode('UTF-8', 'strict')
        finally:
            sundown.bufrelease(ob)


cdef class BaseRenderer:
    """The ``BaseRenderer`` is boilerplate code for creating your own renderers by
    sublassing `BaseRenderer`. It takes care of setting the callbacks and flags.

    :param flags: Available as a read-only, integer type attribute named ``self.flags``.
    """

    cdef sundown.sd_callbacks callbacks
    cdef wrapper.renderopt options

    #: Read-only render flags
    cdef readonly int flags

    def __cinit__(self, int flags=0):
        self.options.self = <void *> self
        self.flags = flags
        self.setup()

        # Set callbacks
        cdef void **source = <void **> &wrapper.callback_funcs
        cdef void **dest = <void **> &self.callbacks

        cdef unicode method_name
        for i from 0 <= i < <int> wrapper.method_count by 1:
            # In Python 3 ``wrapper.method_names[i]`` is a byte string.
            # This means hasattr can't find any method in the renderer, so
            # ``wrapper.method_names[i]`` is converted to a normal string first.
            method_name = wrapper.method_names[i].decode('utf-8')
            if hasattr(self, method_name):
                dest[i] = source[i]

    def setup(self):
        """A method that can be overridden by the renderer that sublasses ``BaseRenderer``.
        It's called everytime an instance of a renderer is created.
        """
        pass


cdef class HtmlRenderer(BaseRenderer):
    """The HTML renderer that's included in Sundown.

    Do you override the ``setup`` method when subclassing ``HtmlRenderer``. If
    you do make sure to call parent class' ``setup`` method first.

    :param flags: Adjust HTML rendering behaviour with the ``HTML_*`` constants.
    """
    def setup(self):
        self.options.html.flags = self.flags
        sundown.sdhtml_renderer(
            &self.callbacks,
            &self.options.html,
            self.options.html.flags)


cdef class HtmlTocRenderer(BaseRenderer):
    """The HTML table of contents renderer that's included in Sundown.

    Do you override the ``setup`` method when subclassing ``HtmlTocRenderer``.
    If you do make sure to call parent class' ``setup`` method first.

    :param flags: Adjust HTML rendering behaviour with the ``HTML_*`` constants.
    """
    def setup(self, int flags=0):
        sundown.sdhtml_toc_renderer(
            &self.callbacks,
            &self.options.html)


cdef class Markdown:
    """The Markdown parser.

    :param renderer: An instance of ``BaseRenderer``.
    :param extensions: Enable additional Markdown extensions with the ``EXT_*`` constants.
    """

    cdef sundown.sd_markdown *markdown
    cdef BaseRenderer renderer

    def __cinit__(self, object renderer, int extensions=0):
        if not isinstance(renderer, BaseRenderer):
            raise ValueError('expected instance of BaseRenderer, %s found' % \
                renderer.__class__.__name__)

        self.renderer = <BaseRenderer> renderer
        self.markdown = sundown.sd_markdown_new(
            extensions, 16,
            &self.renderer.callbacks,
            <sundown.html_renderopt *> &self.renderer.options)

    def render(self, object text):
        """Render the Markdon text.

        Returns a unicode string.

        :param text: A byte or unicode string.
        """
        if hasattr(self.renderer, 'preprocess'):
            text = self.renderer.preprocess(text)

        # Convert string
        cdef bytes py_string
        if hasattr(text, 'encode'):
            py_string = text.encode('UTF-8', 'strict')
        else:
            py_string = text
        cdef char *c_string = py_string

        # Buffers
        cdef sundown.buf *ib = sundown.bufnew(128)
        sundown.bufputs(ib, c_string)

        cdef sundown.buf *ob = sundown.bufnew(128)
        sundown.bufgrow(ob, <size_t> (ib.size * 1.4))

        # Parse! And make a unicode string
        sundown.sd_markdown_render(ob, ib.data, ib.size, self.markdown)
        text = (<char *> ob.data)[:ob.size].decode('UTF-8', 'strict')

        if hasattr(self.renderer, 'postprocess'):
            text = self.renderer.postprocess(text)

        # Return a string and release buffers
        try:
            return text
        finally:
            sundown.bufrelease(ob)
            sundown.bufrelease(ib)

    def __dealloc__(self):
        if self.markdown is not NULL:
            sundown.sd_markdown_free(self.markdown)



from libc.stdint cimport uint8_t


cdef extern from 'sundown/buffer.h':
    struct buf:
        uint8_t *data
        size_t size
        size_t asize
        size_t unit

    buf* bufnew(size_t)
    int bufgrow(buf *, size_t)
    void bufcstr(buf *)
    void bufrelease(buf *)
    void bufputs(buf *, char *)


cdef extern from 'sundown/html.h':
    struct _toc_data_st:
        int header_count
        int current_level

    struct html_renderopt:
        _toc_data_st toc_data
        unsigned int flags
        void (*link_attributes)(buf *ob, buf *url, void *self)

    void sdhtml_renderer(
        sd_callbacks *callbacks,
        html_renderopt *options_ptr,
        unsigned int render_flags)
    void sdhtml_toc_renderer(
        sd_callbacks *callbacks,
        html_renderopt *options_ptr)
    void sdhtml_smartypants(
        buf *ob,
        uint8_t *text,
        size_t size)


cdef extern from 'sundown/markdown.h':
    enum mkd_autolink:
        pass

    struct sd_callbacks:
        # Block level callbacks - NULL skips the block
        void (*blockcode)(buf *ob, buf *text, buf *lang, void *opaque)
        void (*blockquote)(buf *ob, buf *text, void *opaque)
        void (*blockhtml)(buf *ob, buf *text, void *opaque)
        void (*header)(buf *ob, buf *text, int level, void *opaque)
        void (*hrule)(buf *ob, void *opaque)
        void (*list)(buf *ob, buf *text, int flags, void *opaque)
        void (*listitem)(buf *ob, buf *text, int flags, void *opaque)
        void (*paragraph)(buf *ob, buf *text, void *opaque)
        void (*table)(buf *ob, buf *header, buf *body, void *opaque)
        void (*table_row)(buf *ob, buf *text, void *opaque)
        void (*table_cell)(buf *ob, buf *text, int flags, void *opaque)

        # Span level callbacks - NULL or return 0 prints the span verbatim
        int (*autolink)(buf *ob, buf *link, mkd_autolink type, void *opaque)
        int (*codespan)(buf *ob, buf *text, void *opaque)
        int (*double_emphasis)(buf *ob, buf *text, void *opaque)
        int (*emphasis)(buf *ob, buf *text, void *opaque)
        int (*image)(buf *ob, buf *link, buf *title, buf *alt, void *opaque)
        int (*linebreak)(buf *ob, void *opaque)
        int (*link)(buf *ob, buf *link, buf *title, buf *content, void *opaque)
        int (*raw_html_tag)(buf *ob, buf *tag, void *opaque)
        int (*triple_emphasis)(buf *ob, buf *text, void *opaque)
        int (*strikethrough)(buf *ob, buf *text, void *opaque)
        int (*superscript)(buf *ob, buf *text, void *opaque)

        # Low level callbacks - NULL copies input directly into the output
        void (*entity)(buf *ob, buf *entity, void *opaque)
        void (*normal_text)(buf *ob, buf *text, void *opaque)

        # Header and footer
        void (*doc_header)(buf *ob, void *opaque)
        void (*doc_footer)(buf *ob, void *opaque)

    enum mkd_autolink:
        pass

    struct sd_markdown:
        pass

    sd_markdown *sd_markdown_new(
        unsigned int extensions,
        size_t max_nesting,
        sd_callbacks *callbacks,
        html_renderopt *opaque)
    void sd_markdown_render(
        buf *ob,
        uint8_t *document,
        size_t doc_size,
        sd_markdown *md)
    void sd_markdown_free(sd_markdown *md)
    void sd_version(int *major, int *minor, int *revision)



import os
import sys
import glob
import shutil
import os.path

try:
    from setuptools import setup, Extension, Command
except ImportError:
    from distutils.core import setup, Extension, Command


dirname = os.path.dirname(os.path.abspath(__file__))


class BaseCommand(Command):
    user_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass


class CleanCommand(BaseCommand):
    description = 'cleanup directories created by packaging and build processes'
    def run(self):
        for path in ['build', 'dist', 'misaka.egg-info', 'docs/_build']:
            if os.path.exists(path):
                path = os.path.join(dirname, path)
                print('removing %s' % path)
                shutil.rmtree(path)


class CythonCommand(BaseCommand):
    description = 'compile Cython files(s) into C file(s)'
    def run(self):
        try:
            from Cython.Compiler.Main import compile
            path = os.path.join(dirname, 'src', 'misaka.pyx')
            print('compiling %s' % path)
            compile(path)
        except ImportError:
            print('Cython is not installed. Please install Cython first.')


class VendorCommand(BaseCommand):
    description = 'update Sundown files. Use `git submodule foreach git pull` to the most recent files'
    def run(self):
        files = []
        dest = os.path.join(dirname, 'src/sundown')

        for path in ['vendor/sundown/src/*', 'vendor/sundown/html/*']:
            files += glob.glob(os.path.join(dirname, path))

        for path in files:
            if os.path.exists(path):
                print('copy %s -> %s' % (path, dest))
                shutil.copy(path, dest)


setup(
    name='misaka',
    version='1.0.2',
    description='The Python binding for Sundown, a markdown parsing library.',
    author='Frank Smit',
    author_email='frank@61924.nl',
    url='http://misaka.61924.nl/',
    license='MIT',
    long_description=open(os.path.join(dirname, 'README.rst')).read(),
    scripts=['scripts/misaka'],
    cmdclass={
        'clean': CleanCommand,
        'compile_cython': CythonCommand,
        'update_vendor': VendorCommand
    },
    ext_modules=[Extension('misaka', [
        'src/misaka.c',
        'src/wrapper.c',
        'src/sundown/stack.c',
        'src/sundown/buffer.c',
        'src/sundown/markdown.c',
        'src/sundown/html.c',
        'src/sundown/html_smartypants.c',
        'src/sundown/houdini_href_e.c',
        'src/sundown/houdini_html_e.c',
        'src/sundown/autolink.c'
    ])],
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: C',
        'Programming Language :: Cython',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Topic :: Text Processing :: Markup',
        'Topic :: Text Processing :: Markup :: HTML',
        'Topic :: Utilities'
    ]
)



from sundown cimport buf, html_renderopt, mkd_autolink, sd_callbacks


cdef extern from *:
    ctypedef char* const_char_ptr "const char *"
    ctypedef char* const_size_t "const size_t"


cdef extern from 'wrapper.h':
    struct renderopt:
        html_renderopt html
        void *self

    sd_callbacks callback_funcs
    const_char_ptr method_names[]
    const_size_t method_count
